"""Plottings function for Spatio-temporal data."""
import matplotlib.pyplot as plt
import numpy as np

from .utils import array2zigzag


def carpet_plot(
    arr,
    unfold="classic",
    colorbar=True,
    transpose=False,
    figsize=None,
):  # pragma: no cover
    """Carpet Plot of ND array.

    The last dimension if plotted on the horizontal axis,
    The other are flattened on the vertical one.

    Parameters
    ----------
    arr: numpy.ndarray
        The  ND array to plot.
    unfold: str
        The method to unfold the (N-1)D first dimensions.
        If "classic" the data is simply row-wise unfolded
        If "zigzag" (only for 2D spatial dimensions), the data is unfolded following
        a diagonal zig-zag pattern starting from top left.
    colorbar: bool
        flag to show a colorbar
    transpose: bool
        If true, switch the horizontal and vertical axis.

    Returns
    -------
    matplotlib.pyplot.figure:
        The plotted figure.
    """
    if unfold == "classic":
        unfolded_arr = np.reshape(arr, (-1, arr.shape[-1]))
    elif unfold == "zigzag":
        if arr.ndim != 3:
            raise ValueError(
                "input array should be 3 dimensional, "
                "with temporal dimension in last position."
            )
        unfolded_arr = array2zigzag(arr)
    else:
        raise ValueError(f"invalid unfold method {unfold}")

    if np.iscomplexobj(unfolded_arr):
        unfolded_arr = np.abs(unfolded_arr)

    fig, ax = plt.subplots(figsize=figsize)

    if transpose:
        im = ax.imshow(unfolded_arr.T)
        ax.set_ylabel("time")
        ax.set_xlabel("voxel")
    else:
        im = ax.imshow(unfolded_arr)
        ax.set_xlabel("time")
        ax.set_ylabel("voxel")
    if colorbar:
        fig.colorbar(im)
    return fig



"""Shepp-Logan phantom for use with MR simulations.

From https://github.com/mckib2/phantominator/blob/master/phantominator/mr_shepp_logan.py
"""

import numpy as np


def mr_shepp_logan_t2_star(N, B0=3):
    """Return a 3D T2*-weighted Shepp-Logan phantom."""
    return mr_shepp_logan(N, E=None, B0=B0, T2star=True)[-1]


def mr_shepp_logan(N, E=None, B0=3, T2star=False, zlims=(-1, 1)):
    """Shepp-Logan phantom with MR tissue parameters.

    Parameters
    ----------
    N : int or array_like
        Matrix size, (N, N, N), or (L, M, N).
    E : array_like, optional
        ex13 numeric matrix defining e ellipses.  The columns of E
        are:

            - x-coordinate of the center of the ellipsoid (in [-1, 1])
            - y-coordinate of the center of the ellipsoid (in [-1, 1])
            - z-coordinate of the center of the ellipsoid (in [-1, 1])
            - x principal axis of the ellipsoid
            - y principal axis of the ellipsoid
            - z principal axis of the ellipsoid
            - Angle of the ellipsoid (in rad)
            - spin density, M0
            - Parameter A for T1 determination
            - Parameter C for T1 determination
            - Explicit T1 value (in sec, or np.nan if model is used)
            - T2 value (in sec)
            - chi (change in magnetic susceptibility)

        If spin density is negative, M0, T1, and T2 will be subtracted
        instead of cummulatively added.
    B0 : float, optimal
        Field strength (in Tesla).
    T2star : bool, optional
        Use magnetic susceptibility values to return T2star values
        instead of T2. Gyromagnetic ratio is assumed to be that of
        hydrogen.
    zlims : tuple, optional
        Only for 3D.  Specify bounds along z.  Often we only want the
        middle portion of a 3D phantom, e.g., zlim=(-.5, .5).

    Returns
    -------
    M0 : array_like
        The proton density.
    T1 : array_like
        The T1 values.
    T2 : array_like
        The T2 values. If T2star is True, then these will be T2 star
        values.

    Notes
    -----
    Implements the phantoms described in [1]_.

    If parameters A, C are given and T1 is None, T1 is determined
    according to the equation:

        T1 = A*B0^C

    The original source code [2]_

    References
    ----------
    .. [1] Gach, H. Michael, Costin Tanase, and Fernando Boada.
           "2D & 3D Shepp-Logan phantom standards for MRI." 2008 19th
           International Conference on Systems Engineering. IEEE,
           2008.
    .. [2] https://github.com/mckib2/phantominator/blob/master/phantominator \
    /mr_shepp_logan.py
    """
    # Determine size of phantom
    if np.isscalar(N):
        L, M, N = N, N, N
    else:
        L, M, N = N[:]

    # Make sure zlims are appropriate
    assert len(zlims) == 2, (
        "zlims must be a tuple with 2 entries: upper and lower " "bounds!"
    )
    assert zlims[0] <= zlims[1], "zlims: lower bound must be first entry!"

    # Get parameters from paper if None provided
    if E is None:
        E = mr_ellipsoid_parameters()

    # Extract some parameters so we can use them
    xs = E[:, 0]
    ys = E[:, 1]
    zs = E[:, 2]
    xaxis = E[:, 3]
    yaxis = E[:, 4]
    zaxis = E[:, 5]
    theta = E[:, 6]
    M0 = E[:, 7]
    As = E[:, 8]
    Cs = E[:, 9]
    T1 = E[:, 10]
    T2 = E[:, 11]
    chis = E[:, 12]

    # Initialize array
    X, Y, Z = np.meshgrid(  # meshgrid does X, Y backwards
        np.linspace(-1, 1, M), np.linspace(-1, 1, L), np.linspace(zlims[0], zlims[1], N)
    )
    ct = np.cos(theta)
    st = np.sin(theta)
    sgn = np.sign(M0)
    T1s = np.zeros((L, M, N))
    T2s = np.zeros((L, M, N))
    M0s = np.zeros((L, M, N))

    # We'll need the gyromagnetic ratio if returning T2star values
    if T2star:
        # see https://en.wikipedia.org/wiki/Gyromagnetic_ratio:
        gamma0 = 267.52219  # 10^6 rad⋅s−1⋅T⋅−1

    # Put ellipses where they need to be
    for ii in range(E.shape[0]):
        xc, yc, zc = xs[ii], ys[ii], zs[ii]
        a, b, c = xaxis[ii], yaxis[ii], zaxis[ii]
        ct0, st0 = ct[ii], st[ii]

        # Find indices falling inside the ellipsoid, ellipses only
        # rotated in xy plane
        idx = ((X - xc) * ct0 + (Y - yc) * st0) ** 2 / a**2 + (
            (X - xc) * st0 - (Y - yc) * ct0
        ) ** 2 / b**2 + (Z - zc) ** 2 / c**2 <= 1

        # Add ellipses together -- subtract of M0 is negative
        M0s[idx] += M0[ii]

        # Use T2star values if user asked for them
        if T2star:
            T2s[idx] += sgn[ii] / (1 / T2[ii] + gamma0 * np.abs(B0 * chis[ii]))
        else:
            T2s[idx] += sgn[ii] * T2[ii]

        # Use T1 model if not given explicit T1 value
        if np.isnan(T1[ii]):
            T1s[idx] += sgn[ii] * As[ii] * (B0 ** Cs[ii])
        else:
            T1s[idx] += sgn[ii] * T1[ii]

    return (M0s, T1s, T2s)


def mr_ellipsoid_parameters():
    """Return parameters of ellipsoids.

    Returns
    -------
    E : array_like
        Parameters for the ellipsoids used to construct the phantom.
    """
    params = _mr_relaxation_parameters()

    E = np.zeros((15, 13))
    # [:, [x, y, z, a, b, c, theta, m0, A, C, (t1), t2, chi]]
    E[0, :] = [0, 0, 0, 0.72, 0.95, 0.93, 0, 0.8, *params["scalp"]]
    E[1, :] = [0, 0, 0, 0.69, 0.92, 0.9, 0, 0.12, *params["marrow"]]
    E[2, :] = [0, -0.0184, 0, 0.6624, 0.874, 0.88, 0, 0.98, *params["csf"]]
    E[3, :] = [0, -0.0184, 0, 0.6524, 0.864, 0.87, 0, 0.745, *params["gray-matter"]]
    E[4, :] = [-0.22, 0, -0.25, 0.41, 0.16, 0.21, np.deg2rad(-72), 0.98, *params["csf"]]
    E[5, :] = [0.22, 0, -0.25, 0.31, 0.11, 0.22, np.deg2rad(72), 0.98, *params["csf"]]
    E[6, :] = [0, 0.35, -0.25, 0.21, 0.25, 0.35, 0, 0.617, *params["white-matter"]]
    E[7, :] = [0, 0.1, -0.25, 0.046, 0.046, 0.046, 0, 0.95, *params["tumor"]]
    E[8, :] = [-0.08, -0.605, -0.25, 0.046, 0.023, 0.02, 0, 0.95, *params["tumor"]]
    E[9, :] = [
        0.06,
        -0.605,
        -0.25,
        0.046,
        0.023,
        0.02,
        np.deg2rad(-90),
        0.95,
        *params["tumor"],
    ]
    E[10, :] = [0, -0.1, -0.25, 0.046, 0.046, 0.046, 0, 0.95, *params["tumor"]]
    E[11, :] = [0, -0.605, -0.25, 0.023, 0.023, 0.023, 0, 0.95, *params["tumor"]]
    E[12, :] = [
        0.06,
        -0.105,
        0.0625,
        0.056,
        0.04,
        0.1,
        np.deg2rad(-90),
        0.93,
        *params["tumor"],
    ]
    E[13, :] = [0, 0.1, 0.625, 0.056, 0.056, 0.1, 0, 0.98, *params["csf"]]
    E[14, :] = [
        0.56,
        -0.4,
        -0.25,
        0.2,
        0.03,
        0.1,
        np.deg2rad(70),
        0.85,
        *params["blood-clot"],
    ]

    # Need to subtract some ellipses here...
    Eneg = np.zeros(E.shape)
    for ii in range(E.shape[0]):
        # Ellipsoid geometry
        Eneg[ii, :7] = E[ii, :7]

        # Tissue property differs after 4th subtracted ellipsoid
        if ii > 3:
            Eneg[ii, 7:] = E[3, 7:]
        else:
            Eneg[ii, 7:] = E[ii - 1, 7:]

    # Throw out first as we skip this one in the paper's table
    Eneg = Eneg[1:, :]

    # Spin density is negative for subtraction
    Eneg[:, 7] *= -1

    # Paper doesn't use last blood-clot ellipsoid
    E = E[:-1, :]
    Eneg = Eneg[:-1, :]

    # Put both ellipsoid groups together
    E = np.concatenate((E, Eneg), axis=0)

    return E


def _mr_relaxation_parameters():
    """Return MR relaxation parameters for certain tissues.

    Returns
    -------
    params : dict
        Gives entries as [A, C, (t1), t2, chi]

    Notes
    -----
    If t1 is None, the model T1 = A*B0^C will be used.  If t1 is not
    np.nan, then specified t1 will be used.
    """
    # params['tissue-name'] = [A, C, (t1 value if explicit), t2, chi]
    params = dict()
    params["scalp"] = [0.324, 0.137, np.nan, 0.07, -7.5e-6]
    params["marrow"] = [0.533, 0.088, np.nan, 0.05, -8.85e-6]
    params["csf"] = [np.nan, np.nan, 4.2, 1.99, -9e-6]
    params["blood-clot"] = [1.35, 0.34, np.nan, 0.2, -9e-6]
    params["gray-matter"] = [0.857, 0.376, np.nan, 0.1, -9e-6]
    params["white-matter"] = [0.583, 0.382, np.nan, 0.08, -9e-6]
    params["tumor"] = [0.926, 0.217, np.nan, 0.1, -9e-6]
    return params


def _hamming1d(n):
    """Compute the 1D Hamming window."""
    return 0.54 - (0.46 * np.cos(np.arange(n) * 2 * np.pi / (n - 1)))


def g_factor_map(volume_shape, window_type="hamming"):
    """
    Return a g-factor map using a window function.

    Parameters
    ----------
    volume_shape: tuple
        The volume shape, it should be 2 or 3 element tuple.
    window_type: "hamming"
        other type not implemented yet.
    """
    if window_type != "hamming":
        raise NotImplementedError

    window = _hamming1d

    w1 = window(volume_shape[0])
    w2 = window(volume_shape[1])
    w1 = w1 - min(w1) + 1
    w2 = w2 - min(w2) + 1
    g_map = np.outer(w1, w2)

    if len(volume_shape) == 3:
        w3 = window(volume_shape[2])
        w3 = w3 - min(w3) + 1
        g_map = g_map[..., np.newaxis] * w3[np.newaxis, np.newaxis, :]

    return g_map



"""Test for the different denoising methods."""
import numpy as np
import numpy.testing as npt
import pytest

from patch_denoise.denoise import (
    adaptive_thresholding,
    hybrid_pca,
    mp_pca,
    nordic,
    optimal_thresholding,
    raw_svt,
)

from patch_denoise.simulation.phantom import g_factor_map


@pytest.mark.parametrize("recombination", ["weighted", "average", "center"])
def test_mppca_denoiser(phantom, noisy_phantom, recombination):
    """Test the MP-PCA denoiser."""
    denoised, weights, noise = mp_pca(
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        threshold_scale=2.3,
        recombination=recombination,
    )
    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


@pytest.mark.parametrize("recombination", ["weighted", "average", "center"])
def test_hybridpca_denoiser(phantom, noisy_phantom, recombination):
    """Test the Hybrid-PCA denoiser."""
    denoised, weights, noise = hybrid_pca(
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        noise_std=1.0,
        recombination=recombination,
    )

    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


@pytest.mark.parametrize("recombination", ["weighted", "average", "center"])
def test_nordic_denoiser(phantom, noisy_phantom, recombination):
    """Test the Hybrid-PCA denoiser."""
    denoised, weights, noise = nordic(
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        noise_std=1.0,
        recombination=recombination,
    )

    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


@pytest.mark.parametrize("recombination", ["weighted", "average", "center"])
def test_rawsvt_denoiser(phantom, noisy_phantom, recombination):
    """Test the Hybrid-PCA denoiser."""
    denoised, weights, noise = raw_svt(
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        threshold=10,
        recombination=recombination,
    )

    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


@pytest.mark.parametrize("recombination", ["weighted", "average", "center"])
@pytest.mark.parametrize("loss", ["fro", "nuc", "ope"])
def test_optimal_denoiser(phantom, noisy_phantom, recombination, loss):
    """Test the Optimal Thresholding denoiser."""
    denoised, weights, noise = optimal_thresholding(
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        recombination=recombination,
        loss=loss,
    )

    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


@pytest.mark.parametrize("recombination", ["weighted", "average", "center"])
@pytest.mark.parametrize("loss", ["fro", "nuc", "ope"])
def test_optimal_denoiser2(phantom, noisy_phantom, recombination, loss):
    """Test the Optimal Thresholding denoiser with noise apriori provided."""
    denoised, weights, noise = optimal_thresholding(
        noisy_phantom,
        patch_shape=10,
        patch_overlap=9,
        noise_std=1.414 * g_factor_map(phantom.shape[:-1]),
        recombination=recombination,
        loss=loss,
    )
    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


# center is not tested, it takes too much time.
@pytest.mark.parametrize("recombination", ["weighted", "average"])
@pytest.mark.parametrize(
    "method, gamma",
    [("qut", None), ("gsure", np.linspace(1, 5, 10)), ("sure", np.linspace(1, 5, 10))],
)
def test_adaptive_denoiser(phantom, noisy_phantom, recombination, method, gamma):
    """Test the Adaptive Thresholding denoiser."""
    denoised, weights, noise = adaptive_thresholding(
        noisy_phantom,
        patch_shape=10,
        patch_overlap=0,
        recombination=recombination,
        method=method,
        noise_std=2 * g_factor_map(phantom.shape[:-1]),
        gamma0=gamma,
        nbsim=500,
    )

    noise_std_before = np.sqrt(np.nanmean(np.nanvar(noisy_phantom - phantom, axis=-1)))
    noise_std_after = np.sqrt(np.nanmean(np.nanvar(denoised - phantom, axis=-1)))
    assert noise_std_after < noise_std_before


def test_raise_nordic(phantom, noisy_phantom):
    """Test raise error for nordic."""
    npt.assert_raises(
        ValueError,
        nordic,
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        noise_std="not_a_float_or_array",
    )



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""
# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
from importlib.metadata import version

sys.path.insert(0, os.path.abspath("../.."))  # Source code dir relative to this file

# -- Project information -----------------------------------------------------

project = "patch-denoise"
copyright = "2022, Pierre-Antoine Comby"
author = "Pierre-Antoine Comby"

release = version("patch-denoise")
# for example take major/minor
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_gallery.gen_gallery",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# generate autosummary even if no references
autosummary_generate = True
# autosummary_imported_members = True
autodoc_inherit_docstrings = True

napoleon_include_private_with_doc = True

# -- Options for Sphinx Gallery ----------------------------------------------

sphinx_gallery_conf = {
    "examples_dirs": ["../examples/"],
    "filename_pattern": "/example_",
    "ignore_pattern": r"/(__init__|conftest)\.py",
    "reference_url": {
        "numpy": "http://docs.scipy.org/doc/numpy-1.9.1",
        "scipy": "http://docs.scipy.org/doc/scipy-0.17.0/reference",
    },
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_context = {"default_mode": "light"}



"""Common utilities for bindings."""
from dataclasses import dataclass

import numpy as np

from patch_denoise.denoise import (
    hybrid_pca,
    mp_pca,
    nordic,
    optimal_thresholding,
    raw_svt,
    adaptive_thresholding,
)

NIBABEL_AVAILABLE = True
try:
    import nibabel as nib
except ImportError:
    NIBABEL_AVAILABLE = False

DENOISER_MAP = {
    None: None,
    "mp-pca": mp_pca,
    "hybrid-pca": hybrid_pca,
    "raw": raw_svt,
    "optimal-fro": lambda *args, **kwargs: optimal_thresholding(
        *args, loss="fro", **kwargs
    ),
    "optimal-fro-noise": lambda *args, **kwargs: optimal_thresholding(
        *args, loss="fro", **kwargs
    ),
    "optimal-nuc": lambda *args, **kwargs: optimal_thresholding(
        *args, loss="nuc", **kwargs
    ),
    "optimal-ope": lambda *args, **kwargs: optimal_thresholding(
        *args, loss="ope", **kwargs
    ),
    "nordic": nordic,
    "adaptive-qut": lambda *args, **kwargs: adaptive_thresholding(
        *args, method="qut", **kwargs
    ),
}

_RECOMBINATION = {"w": "weighted", "c": "center", "a": "average"}


@dataclass
class DenoiseParameters:
    """Denoise Parameters data structure."""

    method: str = None
    patch_shape: int = 11
    patch_overlap: int = 0
    recombination: str = "weighted"  # "center" is also available
    mask_threshold: int = 10

    @classmethod
    def from_str(self, config_str):
        """Parse config string to create data structure.

        The full format is ::

            <method>_<patch_shape>_<patch_overlap>_<recombination>_<mask_threshold>
        """
        if "noisy" in config_str:
            return DenoiseParameters(
                method=None,
                patch_shape=None,
                patch_overlap=None,
                recombination=None,
                mask_threshold=None,
            )
        else:
            conf = config_str.split("_")
            d = DenoiseParameters()
            if conf:
                d.method = conf.pop(0)
            if conf:
                d.patch_shape = int(conf.pop(0))
            if conf:
                d.patch_overlap = int(conf.pop(0))
            if conf:
                c = conf.pop(0)
                try:
                    d.recombination = _RECOMBINATION[c]
                except KeyError as exc:
                    raise ValueError(f"unsupported  recombination key: {c}") from exc
            if conf:
                d.mask_threshold = int(conf.pop(0))
            return d

    def __str__(self):
        """Return string reprensation."""
        ret_str = f"{self.method}_{self.patch_shape}_{self.patch_overlap}_"
        ret_str += self.recombination[0]
        if self.mask_threshold:
            ret_str += f"_{self.mask_threshold}"
        return ret_str


def load_complex_nifti(mag_file, phase_file, filename=None):  # pragma: no cover
    """Load two nifti image (magnitude and phase) to create a complex valued array.

    Optionally, the result can be save as a .npy file

    Parameters
    ----------
    mag_file: str
        The source magnitude file
    phase_file: str
        The source phase file
    filename: str, default None
        The output filename
    """
    if not NIBABEL_AVAILABLE:
        raise RuntimeError(
            "nibabel is not available, please install it to load experimental data"
        )

    mag = nib.load(mag_file).get_fdata()
    phase = nib.load(phase_file).get_fdata()
    print(np.min(phase), np.max(phase))
    print(np.min(mag), np.max(mag))
    img = mag * np.exp(1j * phase)

    if filename is not None:
        np.save(filename, img)
    return img



"""Utilities for space-time denoising."""
import numpy as np
from scipy.integrate import quad
from scipy.linalg import eigh, svd


def svd_analysis(input_data):
    """Return the centered SVD decomposition.

    U, S, Vt and M are compute such that:
    ``X = U @ (S * Vt) + M.``

    Parameters
    ----------
    input_data : numpy.ndarray
        The patch

    Returns
    -------
    u_vec, s_vals, v_vec, mean
    """
    mean = np.mean(input_data, axis=0)
    data_centered = input_data - mean
    # TODO  benchmark svd vs svds and order of data.
    u_vec, s_vals, v_vec = svd(data_centered, full_matrices=False)

    return u_vec, s_vals, v_vec, mean


def svd_synthesis(u_vec, s_vals, v_vec, mean, idx):
    """
    Reconstruct ``X = (U @ (S * V)) + M`` with only the max_idx greatest component.

    U, S, V must be sorted in decreasing order.

    Parameters
    ----------
    u_vec : numpy.ndarray
    s_vals : numpy.ndarray
    v_vec : numpy.ndarray
    mean : numpy.ndarray
    idx : int

    Returns
    -------
    np.ndarray: The reconstructed matrix.
    """
    return (u_vec[:, :idx] @ (s_vals[:idx, None] * v_vec[:idx, :])) + mean


def eig_analysis(input_data, max_eig_val=10):
    """
    Return the eigen values and vectors of the autocorrelation of the patch.

    This method is faster than the svd, but the eigen values
    are in increasing order.

    Parameters
    ----------
    input_data : np.ndarray
        A 2D Array
    max_eig_val : int, optional
       For faster results, only the ``max_eig_val`` biggest eigenvalues are
       computed. default = 10

    Returns
    -------
    A : numpy.ndarray
        The centered patch A = X - M
    d : numpy.ndarray
        The eigenvalues of A^H A
    W : numpy.ndarray
        The eigenvector matrix of A^H A
    M : numpy.ndarray
        The mean of the patch along the time axis
    """
    mean = np.mean(input_data, axis=0)
    data_centered = input_data - mean
    M, N = data_centered.shape
    eig_vals, eig_vec = eigh(
        data_centered.conj().T @ data_centered / (M - 1),
        driver="evx",
        subset_by_index=(len(mean) - max_eig_val, len(mean) - 1),
    )

    return data_centered, eig_vals, eig_vec, mean


def eig_synthesis(data_centered, eig_vec, mean, max_val):
    """Reconstruction the denoise patch with truncated eigen decomposition.

    This implements equations (1) and (2) of manjon2013
    """
    eig_vec[:, :-max_val] = 0
    return ((data_centered @ eig_vec) @ eig_vec.conj().T) + mean


def marshenko_pastur_median(beta, eps=1e-7):
    r"""Compute the median of the Marchenko-Pastur Distribution.

    Parameters
    ----------
    beta: float
        aspect ratio of a matrix.
    eps: float
        Precision Parameter
    Return
    ------
    float: the estimated median

    Notes
    -----
    This method Solve :math:`F(x) = 1/2` by dichotomy with
    .. math ::

        F(x) = \int_{\beta_-}^{x} \frac{\sqrt{(\beta_+-t)(t-\beta_-)}}{2\pi\beta t} dt

    The integral is computed using scipy.integrate.quad
    """
    if not (0 <= beta <= 1):
        raise ValueError("Aspect Ratio should be between 0 and 1")

    beta_p = (1 + np.sqrt(beta)) ** 2
    beta_m = (1 - np.sqrt(beta)) ** 2

    def mp_pdf(x):
        """Marchenko Pastur Probability density function."""
        if beta_p >= x >= beta_m:
            return np.sqrt((beta_p - x) * (x - beta_m)) / (2 * np.pi * x * beta)
        else:
            return 0

    change = True
    hibnd = beta_p
    lobnd = beta_m
    # quad return (value, upperbound_error).
    # We only need the integral value

    n = 0
    while change and (hibnd - lobnd) > eps and n < 20:
        change = False
        midpoints = np.linspace(lobnd, hibnd, 5)
        int_estimates = np.array(
            list(map(lambda xx: quad(lambda x: mp_pdf(x), beta_m, xx)[0], midpoints))
        )
        if np.any(int_estimates < 0.5):
            lobnd = np.max(midpoints[int_estimates < 0.5])
            change = True
        if np.any(int_estimates > 0.5):
            hibnd = np.min(midpoints[int_estimates > 0.5])
            change = True
        n += 1
    return (lobnd + hibnd) / 2


def get_patch_locs(p_shape, p_ovl, v_shape):
    """
    Get all the patch top-left corner locations.

    Parameters
    ----------
    vol_shape : tuple
        The volume shape
    patch_shape : tuple
        The patch shape
    patch_overlap : tuple
        The overlap of patch for each dimension.

    Returns
    -------
    numpy.ndarray
        All the patch top-left corner locations.
    """
    # Create an iterator for all the possible patches top-left corner location.
    if len(v_shape) != len(p_shape) or len(v_shape) != len(p_ovl):
        raise ValueError("Dimension mismatch between the arguments.")

    ranges = []
    for v_s, p_s, p_o in zip(v_shape, p_shape, p_ovl):
        if p_o >= p_s:
            raise ValueError(
                "Overlap should be a non-negative integer smaller than patch_size",
            )
        last_idx = v_s - p_s
        range_ = np.arange(0, last_idx, p_s - p_o, dtype=np.int32)
        if range_[-1] < last_idx:
            range_ = np.append(range_, last_idx)
        ranges.append(range_)
    # fast ND-Cartesian product from https://stackoverflow.com/a/11146645
    patch_locs = np.empty(
        [len(arr) for arr in ranges] + [len(p_shape)],
        dtype=np.int32,
    )
    for idx, coords in enumerate(np.ix_(*ranges)):
        patch_locs[..., idx] = coords

    return patch_locs.reshape(-1, len(p_shape))


def estimate_noise(noise_sequence, block_size=1):
    """Estimate the temporal noise standard deviation of a noise only sequence."""
    volume_shape = noise_sequence.shape[:-1]
    noise_map = np.empty(volume_shape)
    patch_shape = (block_size,) * len(volume_shape)
    patch_overlap = (block_size - 1,) * len(volume_shape)

    for patch_tl in get_patch_locs(patch_shape, patch_overlap, volume_shape):
        patch_slice = tuple(
            slice(ptl, ptl + ps) for ptl, ps in zip(patch_tl, patch_shape)
        )
        patch_center_img = tuple(
            slice(ptl + ps // 2, ptl + ps // 2 + 1)
            for ptl, ps in zip(patch_tl, patch_shape)
        )
        noise_map[patch_center_img] = np.std(noise_sequence[patch_slice])
    return noise_map



import pytest
import numpy as np

from patch_denoise.simulation.activations import add_frames
from patch_denoise.simulation.noise import add_temporal_gaussian_noise
from patch_denoise.simulation.phantom import g_factor_map, mr_shepp_logan_t2_star


@pytest.fixture(scope="session")
def rng():
    return np.random.RandomState(42)


@pytest.fixture(scope="session")
def phantom(N_rep=20):
    """Create a dummy phantom with fake activations."""
    return add_frames(mr_shepp_logan_t2_star(64)[32], N_rep)


@pytest.fixture(scope="session")
def noisy_phantom(phantom, rng):
    """Create noisy version of phantom."""
    g_map = g_factor_map(phantom.shape[:-1])
    return add_temporal_gaussian_noise(
        phantom,
        sigma=1,
        rng=rng,
        g_factor_map=g_map,
    )



"""Create Dynamical Data from phantom."""

import numpy as np


def add_frames(volume, n_frames):
    """Add Activation to ground truth volume.

    Parameters
    ----------
    volume: numpy.ndarray
        The Static volume to augment
    n_frames: int
        The number of temporal frames to create

    Returns
    -------
    numpy.ndarray
        The temporal sequence of volume with activations.
    """
    return np.repeat(volume[..., np.newaxis], n_frames, axis=-1)



"""Denoising Operator working on space-time dimension."""
from .base import BaseSpaceTimeDenoiser
from .lowrank import (
    HybridPCADenoiser,
    MPPCADenoiser,
    NordicDenoiser,
    OptimalSVDDenoiser,
    RawSVDDenoiser,
)

__all__ = [
    "BaseSpaceTimeDenoiser",
    "MPPCADenoiser",
    "HybridPCADenoiser",
    "NordicDenoiser",
    "OptimalSVDDenoiser",
    "RawSVDDenoiser",
]



"""Nipype Bindings, Provide bindings To apply patch based denoising."""

import os

import nibabel as nib
import numpy as np
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    File,
    SimpleInterface,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.utils.filemanip import split_filename

from .utils import DENOISER_MAP, DenoiseParameters
from ..space_time.utils import estimate_noise


class PatchDenoiseInputSpec(BaseInterfaceInputSpec):
    """InputSpec for Patch denoising Interface."""

    in_mag = File(
        exists=True,
        xor=["in_real", "in_imag"],
        desc="magnitude input file to denoise.",
    )
    in_real = File(
        exists=True,
        xor=["in_mag"],
        require=["in_imag"],
        desc="Real-part of input file to denoise.",
    )
    in_imag = File(
        exists=True,
        xor=["in_mag"],
        require=["in_real"],
        desc="imaginary part of input file to denoise.",
    )

    mask = File(exists=True)
    noise_std_map = File(desc="noise_std_map")
    denoise_str = traits.Str(desc="string describing the denoiser configuration")
    method = traits.Enum(
        *DENOISER_MAP.keys(),
        xor=["denoise_str"],
        require=["patch_shape", "patch_overlap"],
    )
    patch_shape = traits.Union(
        traits.Int(),
        traits.List(traits.Int(), minlen=3, maxlen=3),
        xor=["denoise_str"],
        require=["denoise_method", "patch_overlap"],
    )
    patch_overlap = traits.Union(
        traits.Int(),
        traits.List(traits.Int(), minlen=3, maxlen=3),
        xor=["denoise_str"],
        require=["patch_shape", "denoise_method"],
    )
    mask_threshold = traits.Int(10)
    recombination = traits.Enum("weighted", "mean")
    extra_kwargs = traits.Dict()


class PatchDenoiseOutputSpec(TraitedSpec):
    """OutputSpec for Denoising Interface."""

    denoised_file = File(desc="denoised file")
    noise_std_map = File(desc="a map of the noise variance.")
    pass


class PatchDenoise(SimpleInterface):
    """Patch based denoising interface."""

    input_spec = PatchDenoiseInputSpec
    output_spec = PatchDenoiseOutputSpec

    _denoise_attrs = [
        "method",
        "patch_shape",
        "patch_overlap",
        "mask_threshold",
        "recombination",
    ]

    def _run_interface(self, runtime):
        # INPUT
        if isdefined(self.inputs.denoise_str):
            d_par = DenoiseParameters.from_str(self.inputs.denoise_str)
        else:
            d_par = DenoiseParameters()
            for attr in PatchDenoise._denoise_attrs:
                setattr(d_par, attr, getattr(self.inputs, attr))

        if isdefined(self.inputs.in_mag):
            data_mag_nii = nib.load(self.inputs.in_mag)
            data = data_mag_nii.get_fdata(dtype=np.float32)
            basename = self.inputs.in_mag
            affine = data_mag_nii.affine
        else:
            data_real_nii = nib.load(self.inputs.in_real)
            affine = data_real_nii.affine
            data_real = data_real_nii.get_fdata(dtype=np.float32)
            data_imag = nib.load(self.inputs.in_imag).get_fdata(dtype=np.float32)
            data = 1j * data_imag
            data += data_real
            basename = self.inputs.in_real

        if isdefined(self.inputs.mask) and self.inputs.mask:
            mask = np.abs(nib.load(self.inputs.mask).get_fdata()) > 0
        else:
            mask = None

        try:
            denoise_func = DENOISER_MAP[d_par.method]
        except KeyError:
            raise ValueError(
                f"unknown denoising denoise_method '{self.inputs.denoise_method}', "
                f"available are {list(DENOISER_MAP.keys())}"
            ) from None

        if isdefined(self.inputs.extra_kwargs) and self.inputs.extra_kwargs:
            extra_kwargs = self.inputs.extra_kwargs
        else:
            extra_kwargs = dict()
        if d_par.method in [
            "nordic",
            "hybrid-pca",
            "adaptive-qut",
            "optimal-fro-noise",
        ]:
            extra_kwargs["noise_std"] = nib.load(self.inputs.noise_std_map).get_fdata()

        if denoise_func is not None:
            # CORE CALL
            denoised_data, _, noise_std_map = denoise_func(
                data,
                patch_shape=d_par.patch_shape,
                patch_overlap=d_par.patch_overlap,
                mask=mask,
                mask_threshold=d_par.mask_threshold,
                recombination=d_par.recombination,
                **extra_kwargs,
            )
        else:
            denoised_data = data
            noise_std_map = np.std(data, axis=-1, dtype=np.float32)
        # OUTPUT
        if np.any(np.iscomplex(denoised_data)):
            denoised_data = np.abs(denoised_data, dtype=np.float32)

        _, base, _ = split_filename(basename)
        base = base.replace("_mag", "")
        base = base.replace("_real", "")
        denoise_filename = f"{base}_d_{d_par.method}.nii"
        noise_map_filename = f"{base}_noise_map.nii"

        denoised_data_img = nib.Nifti1Image(denoised_data, affine=affine)
        denoised_data_img.to_filename(denoise_filename)

        noise_map_img = nib.Nifti1Image(noise_std_map, affine=affine)
        noise_map_img.to_filename(noise_map_filename)

        self._results["denoised_file"] = os.path.abspath(denoise_filename)
        self._results["noise_std_map"] = os.path.abspath(noise_map_filename)

        return runtime


class NoiseStdMapInputSpec(BaseInterfaceInputSpec):
    """InputSpec for Noise Map Estimation."""

    noise_map_file = File(
        exists=True,
        mandatory=True,
        desc="A 0-Volt volume acquisition",
    )
    fft_scale = traits.Int(default=100, desc="scaling parameter of the reconstruction.")
    block_size = traits.Int(default=3, desc="size of spatial block to compute the std.")


class NoiseStdMapOutputSpec(TraitedSpec):
    """OutputSpec for Noise Map Estimation."""

    noise_std_map = File(desc="Spatial variation of noise variance")


class NoiseStdMap(SimpleInterface):
    """Noise std estimation."""

    input_spec = NoiseStdMapInputSpec
    output_spec = NoiseStdMapOutputSpec

    def _run_interface(self, runtime):
        noise_map = nib.load(self.inputs.noise_map_file)
        noise_std_map = estimate_noise(
            noise_map.get_fdata() / self.inputs.fft_scale, self.inputs.block_size
        )
        noise_std_map_img = nib.Nifti1Image(noise_std_map, affine=noise_map.affine)

        filename = os.path.abspath(
            os.path.basename(self.inputs.noise_map_file).split(".")[0] + "_std.nii"
        )
        noise_std_map_img.to_filename(filename)
        self._results["noise_std_map"] = filename

        return runtime



import pytest

from patch_denoise.viz.utils import zigzag2array, array2zigzag, _zigzag
import numpy as np
from numpy.testing import assert_equal


@pytest.mark.parametrize("rows, columns", [(10, 10), (21, 11), (11, 21)])
def test_zigzag(rows, columns):
    arr = np.arange(rows * columns).reshape(rows, columns)

    zigzag = _zigzag(*arr.shape)
    print(tuple(zip(*zigzag)))
    arr_new = zigzag2array(array2zigzag(arr), (rows, columns))

    assert_equal(arr_new, arr)



"""Binding for ModOpt Package."""

import numpy as np
from modopt.opt.proximity import ProximityParent

from .utils import DENOISER_MAP


class LLRDenoiserOperator(ProximityParent):
    """Proximal Operator drop-in replacement using local low rank denoising.

    Parameters
    ----------
    denoiser: str
        name of the denoising method.
    patch_shape: tuple
        The patch shape
    patch_overlap: tuple
        the overlap of each pixel
    mask: numpy.ndarray
        A boolean array, defining a ROI in the volume. Only patch with voxels in the ROI
        will be processed.
    mask_threshold: int
        percentage of the path that should be in the mask in order to be processed.
        If mask_threshold = -1, all the patch are processed, if mask_threshold=100, all
        the voxels of the patch needs to be in the mask
    """

    def __init__(
        self,
        denoiser,
        patch_shape,
        patch_overlap,
        recombination="weighted",
        mask=None,
        mask_threshold=-1,
        progbar=None,
        time_dimension=-1,
        **kwargs,
    ):
        self._denoiser = DENOISER_MAP[denoiser]
        self._params = dict(
            patch_shape=patch_shape,
            patch_overlap=patch_overlap,
            mask=mask,
            mask_threshold=mask_threshold,
            progbar=progbar,
            **kwargs,
        )
        self.op = self._op_method
        self.cost = lambda *args, **kw: np.NaN
        self.time_dimension = time_dimension

    def _op_method(self, data, **kwargs):
        run_kwargs = self._params.copy()
        run_kwargs.update(kwargs)
        return np.moveaxis(
            self._denoiser(
                np.moveaxis(data, self.time_dimension, -1),
                **run_kwargs,
            )[0],
            -1,
            self.time_dimension,
        )



"""Docstring utils.

docdict contains the standard argument documentation found across the package.

Docstring can then use templated argument such as ``$patch_config`` that will be
substitute by their definition (see docdict items).

source:
"""

import inspect
from string import Template


docdict = dict(
    patch_config="""
patch_shape: tuple
    The patch shape
patch_overlap: tuple
    the overlap of each pixel
recombination: str, optional
    The recombination method of the patch. "weighted", "average" or "center".
    default "weighted".""",
    mask_config="""
mask: numpy.ndarray
    A boolean array, defining a ROI in the volume. Only patch with voxels in the ROI
    will be processed.
mask_threshold: int
    percentage of the path that has to be in the mask so that the patch is processed.
    if mask_threshold = -1, all the patch are processed, if mask_threshold=100, all
    the voxels of the patch needs to be in the mask""",
    denoise_return="""
tuple
    numpy.ndarray: The denoised sequence of volume
    numpy.ndarray: The weight of each pixel after the processing.
    numpy.ndarray: If possible, the noise variance distribution in the volume.""",
    input_config="""
input_data: numpy.ndarray
    The input data to denoise. It should be a ND array, and the last
    dimension should a dynamically varying one (eg time).
progbar: tqdm.tqdm Progress bar, optiononal
    An existing Progressbar, default (None) will create a new one.
    """,
    noise_std="""
noise_std: float or numpy.ndarray
    An estimation of the spatial noise map standard deviation.""",
)

# complete the standard config with patch and mask configuration.
docdict["standard_config"] = (
    docdict["input_config"] + docdict["patch_config"] + docdict["mask_config"]
)


def fill_doc(f):
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of. Will be modified in place.

    Returns
    -------
    f : callable
        The function, potentially with an updated ``__doc__``.
    """
    docstring = f.__doc__
    if not docstring:
        return f
    try:
        docstring = Template(inspect.cleandoc(docstring)).safe_substitute(docdict)
        # remove possible gap between headline and body.
        f.__doc__ = docstring.replace("---\n\n", "---\n")
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        raise RuntimeError(f"Error documenting {funcname}s:\n{str(exp)}") from exp
    return f



"""Test space-times utilities."""
from itertools import product

import numpy as np
import pytest

from patch_denoise.space_time.utils import (
    eig_analysis,
    eig_synthesis,
    estimate_noise,
    marshenko_pastur_median,
    svd_analysis,
    svd_synthesis,
)

parametrize_random_matrix = pytest.mark.parametrize(
    "matrix",
    [
        {"cplx": a, "shape": b, "sigma": c}
        for a, b, c in product([False], [(100, 50), (100, 10)], [0.1, 1, 10])
    ],
    indirect=True,
)


@pytest.fixture(scope="function")
def medium_random_matrix(rng):
    """Create random 3D array. with size (200, 200, 100)."""
    shape = (200, 200, 100)
    return rng.randn(*shape)


@pytest.fixture()
def matrix(request):
    """Create random matrix on command with shape and noise level."""
    rng = np.random.RandomState(42)
    sigma = request.param["sigma"]
    M, N = request.param["shape"]
    if not request.param["cplx"]:
        return rng.randn(M, N) * sigma
    return rng.randn(M, N) * sigma + 1j * rng.randn(M, N) * sigma


@pytest.mark.parametrize("beta", np.arange(1, 10) * 0.1)
def test_marshenko_pastur_median(beta, rng, n_runs=10000, n_samples=1000):
    """Test the median estimation of Marshenko Pastur law."""
    print(beta)
    beta_p = (1 + np.sqrt(beta)) ** 2
    beta_m = (1 - np.sqrt(beta)) ** 2

    def f(x):
        """Marchenko Pastur Probability density function."""
        if beta_p >= x >= beta_m:
            return np.sqrt((beta_p - x) * (x - beta_m)) / (2 * np.pi * x * beta)
        else:
            return 0

    integral_median = marshenko_pastur_median(beta, eps=1e-7)

    vals = np.linspace(beta_m, beta_p, n_samples)
    proba = np.array(list(map(f, vals)))
    proba /= np.sum(proba)
    samples = np.zeros(n_runs)
    for i in range(n_runs):
        samples[i] = np.median(rng.choice(vals, size=n_runs, p=proba))
    #    montecarlo_median = np.mean(samples)

    # TODO: increase precision of montecarlo simulation
    assert np.std(samples) <= 0.1 * integral_median


@pytest.mark.parametrize("block_dim", range(5, 10))
def test_noise_estimation(medium_random_matrix, block_dim):
    """Test noise estimation."""
    noise_map = estimate_noise(medium_random_matrix, block_dim)

    real_std = np.nanstd(medium_random_matrix)
    err = np.nanmean(noise_map - real_std)
    assert err <= 0.1 * real_std


@parametrize_random_matrix
def test_svd(matrix):
    """Test SVD functions."""
    U, S, V, M = svd_analysis(matrix)
    new_matrix = svd_synthesis(U, S, V, M, idx=len(S))

    # TODO Refine the precision criteria
    assert np.sqrt(np.mean(np.square(matrix - new_matrix))) <= np.std(matrix) / 10


@parametrize_random_matrix
def test_eig(matrix):
    """Test SVD via eigenvalue decomposition."""
    A, d, W, M = eig_analysis(matrix)
    new_matrix = eig_synthesis(A, W, M, max_val=len(M))
    # TODO Refine the precision criteria
    assert np.sqrt(np.mean(np.square(matrix - new_matrix))) <= np.std(matrix)



"""
Visualisation of spatio-temporal data
=====================================

This is a example script to test various denoising methods on real-word fMRI data.

Source data should a sequence of 2D or 3D data, the temporal dimension being the last one.

"""
import matplotlib.pyplot as plt

from patch_denoise.simulation.phantom import mr_shepp_logan_t2_star, g_factor_map
from patch_denoise.simulation.activations import add_frames
from patch_denoise.simulation.noise import add_temporal_gaussian_noise

# %%
# Setup the parameters for the simulation and noise

SHAPE = (64, 64, 64)
N_FRAMES = 400

NOISE_LEVEL = 2

# %%
# Simulate data for the reconstruction. We use a classical Shepp-Logan phantom, on
# which we add gaussian noise.
# Note that the simulation data is real-valued, for a more pratical case check the
# ``experimental_data.py`` file.

# Create a 2D phantom.
phantom = mr_shepp_logan_t2_star(SHAPE)[32]
ground_truth = add_frames(phantom, N_FRAMES)
g_map = g_factor_map(SHAPE)
print(g_map.shape)

noisy_image = add_temporal_gaussian_noise(ground_truth, sigma=NOISE_LEVEL)

# %%
# Unrolling using carpetplot
# --------------------------
# Any spatio temporal data can be unrolled like a carpet to show the temporal evolution
# of every voxel.

from patch_denoise.viz.plots import carpet_plot


carpet_plot(noisy_image[32:48, 32:48], unfold="classic", transpose=True)

# %%
# In case of 2D data the voxels can be unfolded using a zig-zag pattern.
carpet_plot(noisy_image[32:48, 32:48], unfold="zigzag")



"""Provides a functional entry point for denoising methods."""
from ._docs import fill_doc
from .space_time.lowrank import (
    AdaptiveDenoiser,
    HybridPCADenoiser,
    MPPCADenoiser,
    NordicDenoiser,
    OptimalSVDDenoiser,
    RawSVDDenoiser,
)


@fill_doc
def mp_pca(
    input_data,
    patch_shape,
    patch_overlap,
    mask=None,
    mask_threshold=50,
    recombination="weighted",
    threshold_scale=1.0,
    progbar=None,
):
    """
    Marshenko-Pastur PCA denoising method.

    Parameters
    ----------
    $standard_config
    threshold_scale: float
        An extra factor for the patch denoising.

    Returns
    -------
    $denoise_return

    Notes
    -----
    Follows implementation of [#]_ and the one available in dipy.

    References
    ----------
    .. [#] Manjón, José V., Pierrick Coupé, Luis Concha, Antonio Buades,
           D. Louis Collins, and Montserrat Robles.
           “Diffusion Weighted Image Denoising Using Overcomplete Local PCA.”
           PLOS ONE 8, no. 9 (September 3, 2013): e73021.
           https://doi.org/10.1371/journal.pone.0073021.

    See Also
    --------
    patch_denoise.space_time.lowrank.MPPCADenoiser
    """
    denoiser = MPPCADenoiser(
        patch_shape,
        patch_overlap,
        recombination=recombination,
        threshold_scale=threshold_scale,
    )
    return denoiser.denoise(input_data, mask=mask, mask_threshold=mask_threshold)


@fill_doc
def hybrid_pca(
    input_data,
    patch_shape,
    patch_overlap,
    mask=None,
    mask_threshold=50,
    noise_std=1.0,
    recombination="weighted",
    progbar=None,
):
    """
    Hybrid PCA denoising method.

    Parameters
    ----------
    $standard_config
    $noise_std

    Returns
    -------
    $denoise_return

    Notes
    -----
    Follows implementation of [#]_ .

    References
    ----------
    .. [#]
    https://submissions.mirasmart.com/ISMRM2022/Itinerary/Files/PDFFiles/2688.html

    See Also
    --------
    patch_denoise.space_time.lowrank.HybridPCADenoiser
    """
    denoiser = HybridPCADenoiser(
        patch_shape,
        patch_overlap,
        recombination=recombination,
    )
    return denoiser.denoise(
        input_data,
        mask=mask,
        mask_threshold=mask_threshold,
        noise_std=noise_std,
        progbar=progbar,
    )


@fill_doc
def raw_svt(
    input_data,
    patch_shape,
    patch_overlap,
    mask_threshold=50,
    mask=None,
    threshold=1.0,
    recombination="weighted",
    progbar=None,
):
    """
    Raw singular value thresholding.

    Parameters
    ----------
    $standard_config
    threshold: float
        threshold use for singular value hard thresholding.

    Returns
    -------
    tuple
        numpy.ndarray: The denoised sequence of volume
        numpy.ndarray: The weight of each pixel after the processing.
        numpy.ndarray: If possible, the noise variance distribution in the volume.

    Notes
    -----
    Simple raw hard thresholding of singular values.
    TODO: add support for soft thresholding.

    See Also
    --------
    patch_denoise.space_time.lowrank.MPPCADenoiser
    """
    denoiser = RawSVDDenoiser(
        patch_shape,
        patch_overlap,
        recombination=recombination,
        threshold_value=threshold,
    )
    return denoiser.denoise(
        input_data,
        mask=mask,
        mask_threshold=mask_threshold,
        threshold_scale=1.0,
        progbar=progbar,
    )


@fill_doc
def nordic(
    input_data,
    patch_shape,
    patch_overlap,
    mask_threshold=50,
    mask=None,
    noise_std=1.0,
    recombination="weighted",
    n_iter_threshold=10,
    progbar=None,
):
    """
    NORDIC denoising method.

    Parameters
    ----------
    $standard_config
    $noise_std
    n_iter_threshold: int
        The number of Monte-Carlo Simulation to estimate the global threshold.

    Returns
    -------
    $denoise_return

    Notes
    -----
    Follows implementation of [#]_

    References
    ----------
    .. [#] Moeller, Steen, Pramod Kumar Pisharady, Sudhir Ramanna, Christophe Lenglet,
           Xiaoping Wu, Logan Dowdle, Essa Yacoub, Kamil Uğurbil, and Mehmet Akçakaya.
           “NOise Reduction with DIstribution Corrected (NORDIC) PCA in DMRI with
           Complex-Valued Parameter-Free Locally Low-Rank Processing.”
           NeuroImage 226 (February 1, 2021): 117539.
           https://doi.org/10.1016/j.neuroimage.2020.117539.

    See Also
    --------
    patch_denoise.space_time.lowrank.NordicDenoiser
    """
    denoiser = NordicDenoiser(
        patch_shape,
        patch_overlap,
        recombination=recombination,
    )
    return denoiser.denoise(
        input_data,
        mask=mask,
        mask_threshold=mask_threshold,
        noise_std=noise_std,
        n_iter_threshold=n_iter_threshold,
        progbar=progbar,
    )


@fill_doc
def optimal_thresholding(
    input_data,
    patch_shape,
    patch_overlap,
    mask=None,
    mask_threshold=50,
    loss="fro",
    noise_std=None,
    recombination="weighted",
    eps_marshenko_pastur=1e-7,
    progbar=None,
):
    """
    Optimal thresholing denoising method.

    Parameters
    ----------
    $standard_config
    $noise_std
    loss: str
        The loss for which the optimal thresholding is perform.
    eps_marshenko_pastur: float
        The precision with which the optimal threshold is computed.

    Returns
    -------
    $denoise_return

    Notes
    -----
    Reimplement of the original Matlab code [#]_ in python.

    References
    ----------
    .. [#] Gavish, Matan, and David L. Donoho. “Optimal Shrinkage of Singular Values.”
        IEEE Transactions on Information Theory 63, no. 4 (April 2017): 2137–52.
        https://doi.org/10.1109/TIT.2017.2653801.


    See Also
    --------
    patch_denoise.space_time.lowrank.OptimalSVDDenoiser
    """
    denoiser = OptimalSVDDenoiser(
        patch_shape,
        patch_overlap,
        recombination=recombination,
        loss=loss,
    )
    return denoiser.denoise(
        input_data,
        mask=mask,
        noise_std=noise_std,
        mask_threshold=mask_threshold,
        eps_marshenko_pastur=eps_marshenko_pastur,
        progbar=progbar,
    )


@fill_doc
def adaptive_thresholding(
    input_data,
    patch_shape,
    patch_overlap,
    mask=None,
    mask_threshold=50,
    recombination="weighted",
    method="SURE",
    nbsim=500,
    tau0=None,
    gamma0=None,
    noise_std=1.0,
    progbar=None,
):
    """
    Optimal thresholing denoising method.

    Parameters
    ----------
    $input_config
    $noise_std
    method: str
        The adaptive method to use "SURE" or "GSURE"
    nbsim:
        Number of simulation for computing sure estimator
    tau:
        Simulation parameter.
    gamma0:
        Simulation parameter.

    Returns
    -------
    $denoise_return

    Notes
    -----
    Reimplements the R package [#]_ in python.

    References
    ----------
    .. [#] J. Josse and S. Sardy, “Adaptive Shrinkage of singular values.”
           arXiv, Nov. 22, 2014.
           doi: 10.48550/arXiv.1310.6602.

    See Also
    --------
    patch_denoise.space_time.AdaptiveDenoiser
    """
    denoiser = AdaptiveDenoiser(
        patch_shape,
        patch_overlap,
        recombination=recombination,
        method=method,
        nbsim=nbsim,
    )
    return denoiser.denoise(
        input_data,
        mask,
        mask_threshold,
        tau0,
        noise_std,
        gamma0,
        progbar=progbar,
    )



"""Utility function for vizualisation."""
import numpy as np


def _zigzag(rows, columns):
    """Return a list of coordinate in zigzag pattern."""
    pattern = [[] for _ in range(rows + columns - 1)]
    for i in range(rows):
        for j in range(columns):
            s = i + j
            if s % 2 == 0:
                pattern[s].insert(0, (i, j))
            else:
                pattern[s].append((i, j))

    return [p for pp in pattern for p in pp]


def array2zigzag(array):
    """Flatten a 2D array with a zigzag pattern.

    Parameters
    ----------
    array: numpy.ndarray
        a 2 Array to flatten.

    Returns
    -------
    numpy.ndarray
        A flattened 1D array.

    Notes
    -----
    Flattenning a 2D array with a zigzag pattern preserves a spatial contiguity
    in the flattened version, unlike the classical row or column-wise stacking.
    """
    zigzag = _zigzag(*array.shape[:2])
    return array[tuple(zip(*zigzag))]


def zigzag2array(array, shape):
    """
    Reshape a 1d array into a 2D shape following a zigzag pattern.

    Parameters
    ----------
    array: numpy.ndarray
        A 1D array.
    shape: tuple
        The row and column size of the new array

    Returns
    -------
    numpy.ndarray
        The 2D array
    """
    if len(shape) != 2:
        raise ValueError("The shape of the new array should be 2-dimensional")
    zigzag = _zigzag(*shape)

    new_array = np.zeros(shape, dtype=array.dtype)

    for idx, z in enumerate(zigzag):
        print(idx, z)
        new_array[z[0], z[1]] = array[idx]
    return new_array



"""
Experimental Data denoising
===========================

This is a example script to test various denoising methods on real-word fMRI data.

Source data should a sequence of 2D or 3D data, the temporal dimension being the last one.

The available denoising methods are "nordic", "mp-pca", "hybrid-pca", "opt-fro", "opt-nuc" and "opt-op".
"""

from patch_denoise.simulation.phantom import mr_shepp_logan_t2_star, g_factor_map
from patch_denoise.simulation.activations import add_frames
from patch_denoise.simulation.noise import add_temporal_gaussian_noise

# %%
# Setup the parameters for the simulation and noise

SHAPE = (64, 64, 64)
N_FRAMES = 200

NOISE_LEVEL = 2



"""Bindings for different external packages.

3 Bindings are available:

* A `Nipype`<https://github.com/nipype/nipype/>_ binding
* A `Modopt`<https://cea-cosmic.github.io/ModOpt/index.html>_ binding
* A Command line interface, available as ``patch-denoise`` after installation.
"""



"""Low Rank Denoising methods."""
from types import MappingProxyType

import numpy as np
from scipy.linalg import svd
from scipy.optimize import minimize

from .base import BaseSpaceTimeDenoiser
from .utils import (
    eig_analysis,
    eig_synthesis,
    marshenko_pastur_median,
    svd_analysis,
    svd_synthesis,
)
from .._docs import fill_doc

NUMBA_AVAILABLE = True
try:
    import numba as nb
except ImportError:
    NUMBA_AVAILABLE = False
    pass


@fill_doc
class MPPCADenoiser(BaseSpaceTimeDenoiser):
    """Denoising using the MP-PCA threshoding.

    Parameters
    ----------
    $patch_config
    threshold_scale: float
        An extra factor multiplying the threshold.
    """

    def __init__(self, patch_shape, patch_overlap, threshold_scale, **kwargs):
        super().__init__(patch_shape, patch_overlap, **kwargs)
        self.input_denoising_kwargs["threshold_scale"] = threshold_scale

    def _patch_processing(self, patch, patch_slice=None, threshold_scale=1.0):
        """Process a pach with the MP-PCA method."""
        p_center, eig_vals, eig_vec, p_tmean = eig_analysis(patch)
        maxidx = 0
        meanvar = np.mean(eig_vals)
        meanvar *= 4 * np.sqrt((len(eig_vals) - maxidx + 1) / len(patch))
        while meanvar < eig_vals[~maxidx] - eig_vals[0]:
            maxidx += 1
            meanvar = np.mean(eig_vals[:-maxidx])
            meanvar *= 4 * np.sqrt((len(eig_vec) - maxidx + 1) / len(patch))
        var_noise = np.mean(eig_vals[: len(eig_vals) - maxidx])

        maxidx = np.sum(eig_vals > (var_noise * threshold_scale**2))

        if maxidx == 0:
            patch_new = np.zeros_like(patch) + p_tmean
        else:
            patch_new = eig_synthesis(p_center, eig_vec, p_tmean, maxidx)

        # Equation (3) of Manjon 2013
        weights = 1.0 / (1.0 + maxidx)
        noise_map = var_noise * weights
        patch_new *= weights

        return patch_new, noise_map, weights


@fill_doc
class HybridPCADenoiser(BaseSpaceTimeDenoiser):
    """Denoising using the Hybrid-PCA thresholding method.

    Parameters
    ----------
    $patch_config
    """

    @fill_doc
    def denoise(
        self, input_data, mask=None, mask_threshold=50, noise_std=1.0, progbar=None
    ):
        """Denoise using the Hybrid-PCA method.

        Along with the input data a noise std map or value should be provided.

        Parameters
        ----------
        $input_config
        $mask_config
        $noise_std

        Returns
        -------
        $denoise_return
        """
        if isinstance(noise_std, (float, np.floating)):
            self.input_denoising_kwargs["var_apriori"] = noise_std**2 * np.ones(
                input_data.shape[:-1]
            )
        else:
            self.input_denoising_kwargs["var_apriori"] = noise_std**2

        return super().denoise(input_data, mask, mask_threshold, progbar=progbar)

    def _patch_processing(self, patch, patch_slice=None, var_apriori=None):
        """Process a pach with the Hybrid-PCA method."""
        varest = np.mean(var_apriori[patch_slice])
        p_center, eig_vals, eig_vec, p_tmean = eig_analysis(patch)
        maxidx = 0
        var_noise = np.mean(eig_vals)
        while var_noise > varest and maxidx < len(eig_vals) - 2:
            maxidx += 1
            var_noise = np.mean(eig_vals[:-maxidx])
        if maxidx == 0:  # all eigen values are noise
            patch_new = np.zeros_like(patch) + p_tmean
        else:
            patch_new = eig_synthesis(p_center, eig_vec, p_tmean, maxidx)
        # Equation (3) of Manjon2013
        weights = 1.0 / (1.0 + maxidx)
        noise_map = var_noise * weights
        patch_new *= weights

        return patch_new, noise_map, weights


@fill_doc
class RawSVDDenoiser(BaseSpaceTimeDenoiser):
    """
    Classical Patch wise singular value thresholding denoiser.

    Parameters
    ----------
    $patch_config
    threshold_vlue: float
        treshold value for the singular values.
    """

    def __init__(
        self, patch_shape, patch_overlap, threshold_value=1.0, recombination="weighted"
    ):
        self._threshold_val = threshold_value

        super().__init__(patch_shape, patch_overlap, recombination)

    @fill_doc
    def denoise(
        self,
        input_data,
        mask=None,
        mask_threshold=50,
        threshold_scale=1.0,
        progbar=None,
    ):
        """Denoise the input_data, according to mask.

        Patches are extracted sequentially and process by the implemented
        `_patch_processing` function.
        Only patches which have at least a voxel in the mask ROI are processed.

        Parameters
        ----------
        $input_config
        $mask_config
        threshold_scale: float
            Extra factor for the threshold of singular values.

        Returns
        -------
        $denoise_return
        """
        self._threshold = self._threshold_val * threshold_scale
        return super().denoise(input_data, mask, mask_threshold, progbar=progbar)

    def _patch_processing(self, patch, patch_slice=None, **kwargs):
        """Process a pach with the simple SVT method."""
        # Centering for better precision in SVD
        u_vec, s_values, v_vec, p_tmean = svd_analysis(patch)

        maxidx = np.sum(s_values > self._threshold)
        if maxidx == 0:
            p_new = np.zeros_like(patch) + p_tmean
        else:
            s_values[s_values < self._threshold] = 0
            p_new = svd_synthesis(u_vec, s_values, v_vec, p_tmean, maxidx)

        # Equation (3) in Manjon 2013
        theta = 1.0 / (1.0 + maxidx)
        p_new *= theta
        weights = theta

        return p_new, weights, np.NaN


@fill_doc
class NordicDenoiser(RawSVDDenoiser):
    """Denoising using the NORDIC method.

    Parameters
    ----------
    $patch_config
    """

    @fill_doc
    def denoise(
        self,
        input_data,
        mask=None,
        mask_threshold=50,
        noise_std=1.0,
        n_iter_threshold=10,
        progbar=None,
    ):
        """Denoise using the NORDIC method.

        Along with the input data a noise stp map or value should be provided.

        Parameters
        ----------
        $input_config
        $mask_config
        $noise_std

        Returns
        -------
        $denoise_return

        """
        patch_shape, _ = self._BaseSpaceTimeDenoiser__get_patch_param(input_data.shape)
        # compute the threshold using Monte-Carlo Simulations.
        max_sval = sum(
            max(
                svd(
                    np.random.randn(np.prod(patch_shape), input_data.shape[-1]),
                    compute_uv=False,
                )
            )
            for _ in range(n_iter_threshold)
        )
        max_sval /= n_iter_threshold

        if isinstance(noise_std, np.ndarray):
            noise_std = np.mean(noise_std)
        if not isinstance(noise_std, (float, np.floating)):
            raise ValueError(
                "For NORDIC the noise level must be either an"
                + " array or a float specifying the std in the volume.",
            )

        self._threshold = noise_std * max_sval

        return super(RawSVDDenoiser, self).denoise(
            input_data, mask, mask_threshold=mask_threshold, progbar=progbar
        )


# From MATLAB implementation
def _opt_loss_x(y, beta):
    """Compute (8) of donoho2017."""
    tmp = y**2 - beta - 1
    return np.sqrt(0.5 * (tmp + np.sqrt((tmp**2) - (4 * beta)))) * (
        y >= (1 + np.sqrt(beta))
    )


def _opt_ope_shrink(singvals, beta=1):
    """Perform optimal threshold of singular values for operator norm."""
    return np.maximum(_opt_loss_x(singvals, beta), 0)


def _opt_nuc_shrink(singvals, beta=1):
    """Perform optimal threshold of singular values for nuclear norm."""
    tmp = _opt_loss_x(singvals, beta)
    return np.maximum(
        0,
        (tmp**4 - (np.sqrt(beta) * tmp * singvals) - beta),
    ) / ((tmp**2) * singvals)


def _opt_fro_shrink(singvals, beta=1):
    """Perform optimal threshold of singular values for frobenius norm."""
    return np.sqrt(
        np.maximum(
            (((singvals**2) - beta - 1) ** 2 - 4 * beta),
            0,
        )
        / singvals
    )


@fill_doc
class OptimalSVDDenoiser(BaseSpaceTimeDenoiser):
    """
    Optimal Shrinkage of singular values for a specific norm.

    Parameters
    ----------
    $patch_config
    loss: str
        The loss determines the choise of the optimal thresholding function
        associated to it. The losses `"fro"`, `"nuc"` and `"op"` are supported,
        for the frobenius, nuclear and operator norm, respectively.
    """

    _OPT_LOSS_SHRINK = MappingProxyType(
        {
            "fro": _opt_fro_shrink,
            "nuc": _opt_nuc_shrink,
            "ope": _opt_ope_shrink,
        }
    )

    def __init__(
        self,
        patch_shape,
        patch_overlap,
        loss="fro",
        recombination="weighted",
    ):
        super().__init__(patch_shape, patch_overlap, recombination=recombination)
        self.input_denoising_kwargs[
            "shrink_func"
        ] = OptimalSVDDenoiser._OPT_LOSS_SHRINK[loss]

    @fill_doc
    def denoise(
        self,
        input_data,
        mask=None,
        mask_threshold=50,
        noise_std=None,
        eps_marshenko_pastur=1e-7,
        progbar=None,
    ):
        """
        Optimal thresholing denoising method.

        Parameters
        ----------
        $input_config
        $mask_config
        $noise_std
        loss: str
            The loss for which the optimal thresholding is perform.
        eps_marshenko_pastur: float
            The precision with which the optimal threshold is computed.

        Returns
        -------
        $denoise_return

        Notes
        -----
        Reimplement of the original Matlab code [#]_ in python.

        References
        ----------
        .. [#] Gavish, Matan, and David L. Donoho. \
            "Optimal Shrinkage of Singular Values."
            IEEE Transactions on Information Theory 63, no. 4 (April 2017): 2137–52.
            https://doi.org/10.1109/TIT.2017.2653801.
        """
        patch_shape, _ = self._BaseSpaceTimeDenoiser__get_patch_param(input_data.shape)

        self.input_denoising_kwargs["mp_median"] = marshenko_pastur_median(
            beta=input_data.shape[-1] / np.prod(patch_shape),
            eps=eps_marshenko_pastur,
        )
        if noise_std is None:
            self.input_denoising_kwargs["var_apriori"] = None
        elif isinstance(noise_std, (float, np.floating)):
            self.input_denoising_kwargs["var_apriori"] = noise_std**2 * np.ones(
                input_data.shape[:-1]
            )
        else:
            self.input_denoising_kwargs["var_apriori"] = noise_std**2

        return super().denoise(input_data, mask, mask_threshold, progbar=progbar)

    def _patch_processing(
        self,
        patch,
        patch_slice=None,
        shrink_func=None,
        mp_median=None,
        var_apriori=None,
    ):
        u_vec, s_values, v_vec, p_tmean = svd_analysis(patch)
        if var_apriori is not None:
            sigma = np.mean(np.sqrt(var_apriori[patch_slice]))
        else:
            sigma = np.median(s_values) / np.sqrt(patch.shape[1] * mp_median)

        scale_factor = np.sqrt(patch.shape[1]) * sigma
        thresh_s_values = scale_factor * shrink_func(
            s_values / scale_factor,
            beta=patch.shape[1] / patch.shape[0],
        )
        thresh_s_values[np.isnan(thresh_s_values)] = 0

        if np.any(thresh_s_values):
            maxidx = np.max(np.nonzero(thresh_s_values)) + 1
            p_new = svd_synthesis(u_vec, thresh_s_values, v_vec, p_tmean, maxidx)
        else:
            maxidx = 0
            p_new = np.zeros_like(patch) + p_tmean

        # Equation (3) in Manjon 2013
        theta = 1.0 / (1.0 + maxidx)
        p_new *= theta
        weights = theta

        return p_new, weights, np.NaN


def _sure_atn_cost(X, method, sing_vals, gamma, sigma=None, tau=None):
    """
    Compute the SURE cost function.

    Parameters
    ----------
    X: np.ndarray
    sing_vals : singular values of X
    gamma: float
    sigma: float
    tau: float
    """
    n, p = np.shape(X)
    if method == "qut":
        gamma = np.exp(gamma) + 1
    else:
        tau = np.exp(tau)

    sing_vals2 = sing_vals**2
    n_vals = len(sing_vals)
    D = np.zeros((n_vals, n_vals), dtype=np.float32)
    dhat = sing_vals * np.maximum(1 - ((tau / sing_vals) ** gamma), 0)
    tmp = sing_vals * dhat
    for i in range(n_vals):
        diff2i = sing_vals2[i] - sing_vals2
        diff2i[i] = np.inf
        D[i, :] = tmp[i] / diff2i

    gradd = (1 + (gamma - 1) * (tau / sing_vals) ** gamma) * (sing_vals >= tau)
    div = np.sum(gradd + abs(n - p) * dhat / sing_vals) + 2 * np.sum(D)

    rss = np.sum((dhat - sing_vals) ** 2)
    if method == "gsure":
        return rss / (1 - div / n / p) ** 2
    return (sigma**2) * ((-n * p) + (2 * div)) + rss


if NUMBA_AVAILABLE:
    s = nb.float32
    d = nb.float64
    sure_atn_cost = nb.njit(
        [
            s(s[:, :], nb.types.unicode_type, s[:], s, s, s),
            s(s[:, :], nb.types.unicode_type, s[:], s, d, d),
        ],
        fastmath=True,
    )(_sure_atn_cost)


def _atn_shrink(singvals, gamma, tau):
    """Adaptive trace norm shrinkage."""
    return singvals * np.maximum(1 - (tau / singvals) ** gamma, 0)


def _get_gamma_tau_qut(patch, sing_vals, stdest, gamma0, nbsim):
    """Estimate gamma and tau using the quantile method."""
    maxd = np.ones(nbsim)
    for i in range(nbsim):
        maxd[i] = np.max(
            svd(
                np.random.randn(*patch.shape) * stdest,
                compute_uv=False,
                overwrite_a=True,
            )
        )
    # auto estimation of tau.
    tau = np.quantile(maxd, 1 - 1 / np.sqrt(np.log(max(*patch.shape))))
    # single value for gamma not provided, estimating it.
    if not isinstance(gamma0, (float, np.floating)):

        def sure_gamma(gamma):
            return _sure_atn_cost(
                X=patch,
                method="qut",
                sing_vals=sing_vals,
                gamma=gamma,
                sigma=stdest,
                tau=tau,
            )

        res_opti = minimize(sure_gamma, 0)
        gamma = np.exp(res_opti.x) + 1
    else:
        gamma = gamma0
    return gamma, tau


def _get_gamma_tau(patch, sing_vals, stdest, method, gamma0, tau0):
    """Estimate gamma and tau."""
    # estimation of tau
    def sure_tau(tau, *args):
        return _sure_atn_cost(*args, tau[0])

    if tau0 is None:
        tau0 = np.log(np.median(sing_vals))
    cost_glob = np.Inf
    for g in gamma0:
        res_opti = minimize(
            lambda x: _sure_atn_cost(
                X=patch,
                method=method,
                gamma=g,  # noqa: B023
                sing_vals=sing_vals,
                sigma=stdest,
                tau=x,
            ),
            tau0,
        )
        # get cost value.
        cost = _sure_atn_cost(
            X=patch,
            method=method,
            gamma=g,
            sing_vals=sing_vals,
            sigma=stdest,
            tau=res_opti.x,
        )
        if cost < cost_glob:
            gamma = g
            tau = np.exp(res_opti.x)
            cost_glob = cost
    return gamma, tau


@fill_doc
class AdaptiveDenoiser(BaseSpaceTimeDenoiser):
    """Adaptive Denoiser.

    Parameters
    ----------
    $patch_config
    """

    _SUPPORTED_METHOD = ["sure", "qut", "gsure"]

    def __init__(
        self,
        patch_shape,
        patch_overlap,
        method="SURE",
        recombination="weighted",
        nbsim=500,
    ):
        super().__init__(patch_shape, patch_overlap, recombination)
        if method.lower() not in self._SUPPORTED_METHOD:
            raise ValueError(
                f"Unsupported method: '{method}', use any of {self._SUPPORTED_METHOD}"
            )
        self.input_denoising_kwargs["method"] = method.lower()
        self.input_denoising_kwargs["nbsim"] = nbsim

    @fill_doc
    def denoise(
        self,
        input_data,
        mask=None,
        mask_threshold=50,
        tau0=None,
        noise_std=None,
        gamma0=None,
        progbar=None,
    ):
        """
        Adaptive denoiser.

        Perform the denoising using the adaptive trace norm estimator. [#]_

        Parameters
        ----------
        $input_config
        $mask_config
        $noise_std


        References
        ----------
        .. [#] J. Josse and S. Sardy, “Adaptive Shrinkage of singular values.”
            arXiv, Nov. 22, 2014.
            doi: 10.48550/arXiv.1310.6602.


        """
        self.input_denoising_kwargs["gamma0"] = gamma0
        self.input_denoising_kwargs["tau0"] = tau0

        if isinstance(noise_std, (float, np.floating)):
            self.input_denoising_kwargs["var_apriori"] = noise_std**2 * np.ones(
                input_data.shape[:-1]
            )
        else:
            self.input_denoising_kwargs["var_apriori"] = noise_std**2
        return super().denoise(input_data, mask, mask_threshold, progbar=progbar)

    def _patch_processing(
        self,
        patch,
        patch_slice=None,
        gamma0=None,
        tau0=None,
        var_apriori=None,
        method=None,
        nbsim=None,
    ):
        stdest = np.sqrt(np.mean(var_apriori[patch_slice]))

        u_vec, sing_vals, v_vec, p_tmean = svd_analysis(patch)

        if method == "qut":
            gamma, tau = _get_gamma_tau_qut(patch, sing_vals, stdest, gamma0, nbsim)
        else:
            gamma, tau = _get_gamma_tau(patch, sing_vals, stdest, method, gamma0, tau0)
        # end of parameter selection
        # Perform thresholding

        thresh_s_values = _atn_shrink(sing_vals, gamma=gamma, tau=tau)
        if np.any(thresh_s_values):
            maxidx = np.max(np.nonzero(thresh_s_values)) + 1
            p_new = svd_synthesis(u_vec, thresh_s_values, v_vec, p_tmean, maxidx)
        else:
            maxidx = 0
            p_new = np.zeros_like(patch) + p_tmean

        # Equation (3) in Manjon 2013
        theta = 1.0 / (1.0 + maxidx)
        p_new *= theta
        weights = theta

        return p_new, weights, np.NaN



"""Collection of patch-based denoising methods."""
from patch_denoise.denoise import (
    mp_pca,
    hybrid_pca,
    optimal_thresholding,
    adaptive_thresholding,
    raw_svt,
    nordic,
)

from patch_denoise.space_time.lowrank import (
    AdaptiveDenoiser,
    HybridPCADenoiser,
    MPPCADenoiser,
    NordicDenoiser,
    OptimalSVDDenoiser,
    RawSVDDenoiser,
)

__all__ = [
    "AdaptiveDenoiser",
    "HybridPCADenoiser",
    "MPPCADenoiser",
    "NordicDenoiser",
    "OptimalSVDDenoiser",
    "RawSVDDenoiser",
    "mp_pca",
    "hybrid_pca",
    "optimal_thresholding",
    "adaptive_thresholding",
    "raw_svt",
    "nordic",
]

try:
    # -- Distribution mode --
    # import from _version.py generated by setuptools_scm during release
    from ._version import version as __version__
except ImportError:
    # -- Source mode --
    # use setuptools_scm to get the current version from src using git
    from setuptools_scm import get_version as _gv
    from os import path as _path

    __version__ = _gv(_path.join(_path.dirname(__file__), _path.pardir))



"""Base Structure for patch-based denoising on spatio-temporal dimension."""
import abc

import numpy as np
from tqdm.auto import tqdm

from .._docs import fill_doc

from .utils import get_patch_locs


@fill_doc
class BaseSpaceTimeDenoiser(abc.ABC):
    """
    Base Class for Patch-based denoising methods for dynamical data.

    Parameters
    ----------
    $patch_config
    """

    def __init__(self, patch_shape, patch_overlap, recombination="weighted"):
        self.p_shape = patch_shape
        self.p_ovl = patch_overlap
        self.recombination = recombination

        self.input_denoising_kwargs = dict()

    @fill_doc
    def denoise(self, input_data, mask=None, mask_threshold=50, progbar=None):
        """Denoise the input_data, according to mask.

        Patches are extracted sequentially and process by the implemented
        `_patch_processing` function.
        Only patches which have at least a voxel in the mask ROI are processed.

        Parameters
        ----------
        $input_config
        $mask_config

        Returns
        -------
        $denoise_return
        """
        data_shape = input_data.shape
        output_data = np.zeros_like(input_data)
        # Create Default mask
        if mask is None:
            process_mask = np.full(data_shape[:-1], True)
        else:
            process_mask = np.copy(mask)

        patch_shape, patch_overlap = self.__get_patch_param(data_shape)
        patch_size = np.prod(patch_shape)

        if self.recombination == "center":
            patch_center = (
                *(slice(ps // 2, ps // 2 + 1) for ps in patch_shape),
                slice(None, None, None),
            )
        patchs_weight = np.zeros(data_shape[:-1], np.float32)
        noise_std_estimate = np.zeros(data_shape[:-1], dtype=np.float32)

        # discard useless patches
        patch_locs = get_patch_locs(patch_shape, patch_overlap, data_shape[:-1])
        get_it = np.zeros(len(patch_locs), dtype=bool)

        for i, patch_tl in enumerate(patch_locs):
            patch_slice = tuple(
                slice(tl, tl + ps) for tl, ps in zip(patch_tl, patch_shape)
            )
            if 100 * np.sum(process_mask[patch_slice]) / patch_size > mask_threshold:
                get_it[i] = True

        print("Denoise {:.2f}% patches".format(100 * np.sum(get_it) / len(patch_locs)))
        patch_locs = np.ascontiguousarray(patch_locs[get_it])

        if progbar is None:
            progbar = tqdm(total=len(patch_locs))
        elif progbar is not False:
            progbar.reset(total=len(patch_locs))

        for patch_tl in patch_locs:
            patch_slice = tuple(
                slice(tl, tl + ps) for tl, ps in zip(patch_tl, patch_shape)
            )
            process_mask[patch_slice] = 1
            # building the casoratti matrix
            patch = np.reshape(input_data[patch_slice], (-1, input_data.shape[-1]))

            # Replace all nan by mean value of patch.
            # FIXME this behaviour should be documented
            # And ideally choosen by the user.

            patch[np.isnan(patch)] = np.mean(patch)
            p_denoise, *extras = self._patch_processing(
                patch,
                patch_slice=patch_slice,
                **self.input_denoising_kwargs,
            )

            p_denoise = np.reshape(p_denoise, (*patch_shape, -1))
            if self.recombination == "center":
                patch_center_img = tuple(
                    ptl + ps // 2 for ptl, ps in zip(patch_tl, patch_shape)
                )
                output_data[patch_center_img] = p_denoise[patch_center]
                patchs_weight[patch_center_img] += extras[0]
                noise_std_estimate[patch_center_img] += extras[1]
            else:
                output_data[patch_slice] += p_denoise
                if self.recombination == "weighted":
                    patchs_weight[patch_slice] += extras[0]
                elif self.recombination == "average":
                    patchs_weight[patch_slice] += 1
            if len(extras) > 1:
                noise_std_estimate[patch_slice] += extras[1]
            if progbar:
                progbar.update()
        # Averaging the overlapping pixels.
        output_data /= patchs_weight[..., None]
        noise_std_estimate /= patchs_weight

        output_data[~process_mask] = 0

        return output_data, patchs_weight, noise_std_estimate

    @abc.abstractmethod
    def _patch_processing(self, patch, patch_slice=None, **kwargs):
        """Process a patch.

        Implemented by child classes.
        """

    def __get_patch_param(self, data_shape):
        """Return tuple for patch_shape and patch_overlap.

        It works from whatever the  input format was (int or list).
        This method also ensure that the patch will provide tall and skinny matrices.
        """
        pp = [None, None]
        for i, attr in enumerate(["p_shape", "p_ovl"]):
            p = getattr(self, attr)
            if isinstance(p, list):
                p = tuple(p)
            elif isinstance(p, (int, np.integer)):
                p = (p,) * (len(data_shape) - 1)
            pp[i] = p
        if np.prod(pp[0]) < data_shape[-1]:
            raise ValueError(
                f"the number of voxel in patch ({np.prod(pp[0])}) is smaller than the"
                f" last dimension ({data_shape[-1]}), this makes an ill-conditioned"
                "matrix for SVD."
            )
        return tuple(pp)



"""Functions to add different source of temporal noise to data."""
import numpy as np
from numpy.random import default_rng


def add_temporal_gaussian_noise(array, sigma=1, g_factor_map=None, rng=None):
    """Add gaussian noise to array.

    Parameters
    ----------
    array: numpy.ndarray
        The noise_free ND-array, where the last dimension is a dynamical one
        (e.g. time)
    sigma: float
        gaussian noise variance
    g_factor_map: numpy.ndarray, optional
        Spatial variation of the noise ((N-1)D array). default is identity.

    Returns
    -------
    numpy.ndarray
        A noisy array
    """
    if rng is None:
        rng = default_rng()
    shape = array.shape

    g_noise = sigma * rng.standard_normal(shape)
    if g_factor_map is None:
        g_factor_map = np.ones(shape[:-1])

    if np.iscomplex(array).any():
        g_noise += 1j * sigma * rng.standard_normal(shape)
    return array + (g_noise * g_factor_map[..., None])


def add_temporal_rician_noise(array, scale=1, rng=None):
    """Add center rician noise to array.

    Parameters
    ----------
    array: numpy.ndarray
        The noise-free array
    sigma: float
        The scale of the Rice distribution

    Notes
    -----
    This function considered centered Rician noise [1]_,
    and thus the noise generated follows a Rayleigh distribution [2]_.

    References
    ----------
    .. [1] https://en.m.wikipedia.org/wiki/Rice_distribution
    .. [2] https://en.m.wikipedia.org/wiki/Rayleigh_distribution
    """
    if rng is None:
        rng = default_rng()
    noise = rng.rayleigh(scale, array.shape)

    return array + noise



"""Configuration for testing the example scripts."""
from pathlib import Path
import runpy
import pytest


def pytest_collect_file(path, parent):
    """Pytest hook.

    Create a collector for the given path, or None if not relevant.
    The new node needs to have the specified parent as parent.
    """
    p = Path(path)
    if p.suffix == ".py" and "example" in p.name:
        return Script.from_parent(parent, path=p, name=p.name)


class Script(pytest.File):
    """Script files collected by pytest."""

    def collect(self):
        """Collect the script as its own item."""
        yield ScriptItem.from_parent(self, name=self.name)


class ScriptItem(pytest.Item):
    """Item script collected by pytest."""

    def runtest(self) -> None:
        """Run the script as a test."""
        runpy.run_path(str(self.path))

    def repr_failure(self, excinfo):
        """Return only the error traceback of the script."""
        excinfo.traceback = excinfo.traceback.cut(path=self.path)
        return super().repr_failure(excinfo)



#!/usr/bin/env python3
"""Cli interface."""


import argparse
import warnings

import numpy as np

from .utils import DENOISER_MAP, DenoiseParameters

NIBABEL_AVAILABLE = True
try:
    import nibabel as nib
except ImportError:
    NIBABEL_AVAILABLE = False


DENOISER_NAMES = ", ".join(d for d in DENOISER_MAP if d)


def parse_args():
    """Parse input arguments."""
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("input_file", help="Input (noisy) file.")
    parser.add_argument(
        "output_file",
        nargs="?",
        default=None,
        help=("Output (denoised) file.\n" "default is d<input_file_name>"),
    )

    parser.add_argument(
        "--conf",
        help=(
            "denoising configuration.\n"
            "Format should be <name>_<patch-size>_<patch-overlap>_<recombination>.\n"
            "See Documentation of ``DenoiseParameter.from_str`` for full specification."
            "Default:\n   optimal-fro_11_5_w\n"
            "Available denoising methods:\n  " + DENOISER_NAMES
        ),
        default="optimal-fro_11_5_w",
    )
    parser.add_argument(
        "--mask",
        default=None,
        help="mask file",
    )
    parser.add_argument(
        "--noise-map",
        default=None,
        help="noise map estimation file",
    )
    parser.add_argument(
        "--output-noise-map",
        default=None,
        help="output noise map estimation file",
    )

    args = parser.parse_args()

    # default value for output.
    if args.output is None:
        input_path = args.input.split("/")
        args.output = "/".join(input_path)[:-1] + "/d" + input_path[-1]


def load_as_array(input):
    """Load a file as a numpy array, and return affine matrix if avaiable."""
    if input is None:
        return None
    if input.endswith(".npy"):
        return np.load(input), None
    elif input.endswith(".nii") or input.endswith(".nii.gz"):
        return nib.load(input).get_fdata(), input.affine
    else:
        raise ValueError("Unsupported file format. use numpy or nifti formats.")


def save_array(data, affine, filename):
    """Save array to file, with affine matrix if required."""
    if filename is None:
        return None

    if filename.endswith(".nii") or filename.endswith(".nii.gz"):
        if affine is None:
            affine = np.eye(len(data.shape))
        nii_img = nib.Nifti1Image(data, affine)
        nii_img.save(filename)
    elif filename.endswith(".npy"):
        np.save(filename, data)

    return filename


def main():
    """Command line entry point."""
    args = parse_args()
    print(args)

    input_data, affine = load_as_array(args.input)
    mask, affine_mask = load_as_array(args.mask)
    noise_map, affine_noise_map = load_as_array(args.noise_map)

    if affine is not None:
        if affine_mask is not None and np.allclose(affine, affine_mask):
            warnings.warn("Affine matrix of input and mask does not match")
        if affine_noise_map is not None and np.allclose(affine, affine_noise_map):
            warnings.warn("Affine matrix of input and noise map does not match")
    d_par = DenoiseParameters.from_str(args.conf)

    denoise_func = DENOISER_MAP[d_par.method]
    extra_kwargs = dict()

    if d_par.method in [
        "nordic",
        "hybrid-pca",
        "adaptive-qut",
        "optimal-fro-noise",
    ]:
        extra_kwargs["noise_std"] = noise_map
        if noise_map is None:
            raise RuntimeError("A noise map must me specified for this method.")
    denoised_data, _, noise_std_map = denoise_func(
        input_data,
        patch_shape=d_par.patch_shape,
        patch_overlap=d_par.patch_overlap,
        mask=mask,
        mask_threshold=d_par.mask_threshold,
        recombination=d_par.recombination,
        **extra_kwargs,
    )

    save_array(denoised_data, affine, args.output)
    save_array(noise_std_map, affine, args.output_noise_map)


if __name__ == "__main__":
    main()



"""Test for the binding module."""
import os
import numpy as np
import numpy.testing as npt
import pytest

MODOPT_AVAILABLE = True
NIPYPE_AVAILABLE = True
try:
    import modopt
except ImportError as e:
    MODOPT_AVAILABLE = False
try:
    import nipype
except ImportError as e:
    NIPYPE_AVAILABLE = False


from patch_denoise.bindings.modopt import LLRDenoiserOperator
from patch_denoise.bindings.utils import DenoiseParameters
from patch_denoise.denoise import mp_pca


def test_modopt(noisy_phantom):
    """test the Modopt Operator."""
    operator = LLRDenoiserOperator(
        "mp-pca",
        patch_shape=6,
        patch_overlap=5,
        threshold_scale=2.3,
        recombination="weighted",
    )

    denoised_modopt = operator.op(noisy_phantom)
    denoised_func, _, _ = mp_pca(
        noisy_phantom,
        patch_shape=6,
        patch_overlap=5,
        threshold_scale=2.3,
        recombination="weighted",
    )

    npt.assert_allclose(denoised_modopt, denoised_func)


def test_entrypoint():
    exit_status = os.system("patch-denoise --help")
    assert exit_status == 0


def test_denoise_param():

    d = DenoiseParameters(
        "optimal-fro",
        11,
        10,
        "weighted",
        10,
    )

    d2 = DenoiseParameters.from_str(str(d))

    assert d2 == d
